from datetime import datetime
from langchain_core.prompts import ChatPromptTemplate


def get_claim_verification_web_search_query_prompt() -> ChatPromptTemplate:
    """
    Get a prompt to generate a single search engine query from a claim to verify.

    Input keys are:
    - `claim`: the claim to verify
    """
    claim_verification_search_query_prompt_template = """Create a single, focused web search engine query to verify this claim:
---
{claim}
---

Strictly output valid JSON (no markdown or other formatting): {{"query": "query"}}"""
    return ChatPromptTemplate.from_template(
        claim_verification_search_query_prompt_template
    )


def get_route_search_type_prompt() -> ChatPromptTemplate:
    route_search_type_prompt_template = """You are an AI assistant that helps determine whether video or text search results would be more appropriate for a user's query. 

Choose video type if the user query involves:
- a physical demonstration
- a visual process
- DIY work or sequences of steps to be executed by a human
- some spatial understanding

Otherwise, choose text type.

Here is the user query, delimited by triple dashes:
---
{query}
---

Strictly output valid JSON (no markdown or other formatting): {{"search_type": "video" | "text"}}
"""
    return ChatPromptTemplate.from_template(route_search_type_prompt_template)


def get_web_search_agent_system_prompt() -> str:
    prompt = """You are a precise research assistant with web search capabilities. Your tasks:
1. Provide accurate, current information
2. Synthesize multi-source information concisely
3. Include citations and maintain objectivity

Focus on authoritative, verifiable sources only.

Skip web search if you are completely certain of information.
Otherwise, perform targeted searches as needed."""
    # ground the LLM in time
    prompt += f"""
Today is {datetime.now().strftime('%Y-%m-%d')}."""
    return prompt


def get_web_search_query_prompt(excluded_queries: list[str] = []) -> ChatPromptTemplate:
    """
    Get a prompt to generate a single search engine query from a user query.

    Input keys are:
    - `query`: the user query to expand
    """
    web_search_query_prompt_template = """Generate appropriate a single web search engine query to maximize the quality of the search results following a user query.

Here is the user query, delimited by triple dashes:
---
{query}
---

IMPORTANT: generate ONLY one query to best cover the topic

Strictly output valid JSON (no markdown or other formatting):
{{"query": "query"}}"""
    # ground the LLM in time
    web_search_query_prompt_template += f"""
Today is {datetime.now().strftime('%Y-%m-%d')}."""
    # add excluded queries to the prompt
    if len(excluded_queries) > 0:
        web_search_query_prompt_template += f"""
Your output MUST NOT include the following queries:
- {"\n- ".join(excluded_queries)}"""
    return ChatPromptTemplate.from_template(web_search_query_prompt_template)


def get_web_search_queries_prompt(
    excluded_queries: list[str] = [],
) -> ChatPromptTemplate:
    """
    Get a prompt to generate a list of search engine queries from a user query.

    Input keys are:
    - `query`: the user query to expand
    """
    web_search_queries_prompt_template = """Generate appropriate web search engine queries to find objective information about this user query:

---
{query}
---

IMPORTANT: 
- generate as many queries as needed to thoroughly cover the topic
- use fewer queries for simple topics, more for complex ones
- each query should target a distinct aspect of the information needed
- avoid redundant queries

Strictly output valid JSON (no markdown or other formatting):
{{"queries": ["query 1", "query 2", ...]}}"""
    # ground the LLM in time
    web_search_queries_prompt_template += f"""
Today is {datetime.now().strftime('%Y-%m-%d')}."""
    # add excluded queries to the prompt
    if len(excluded_queries) > 0:
        web_search_queries_prompt_template += f"""
Your output MUST NOT include the following queries:
- {"\n- ".join(excluded_queries)}"""
    return ChatPromptTemplate.from_template(web_search_queries_prompt_template)



from langgraph.graph import MessagesState, StateGraph, START
from langgraph.prebuilt import ToolNode, tools_condition
import os
import sys

root_dir = os.path.dirname(os.path.dirname(os.path.dirname(os.path.abspath(__file__))))
sys.path.append(root_dir)

from agentic_search.nodes.web import get_web_search_agent_node, get_web_search_tools


def get_search_the_web_react_graph():
    """Get the react graph for searching the web."""
    builder = StateGraph(MessagesState)

    # add nodes to the graph
    builder.add_node(
        "web_search_agent",
        get_web_search_agent_node,
    )
    builder.add_node(
        "tools",
        ToolNode(get_web_search_tools()),
    )

    # add edges to the graph
    builder.add_edge(START, "web_search_agent")
    # here is the branching logic
    builder.add_conditional_edges(
        "web_search_agent",
        # if the last message from the agent is a tool use, continue to the tool node,
        # otherwise continue to the end of the graph (in this case, the agent itself)
        tools_condition,
    )
    # this simple edge, with the conditional edge, is actually what enables the ReAct pattern:
    # this change creates a powerful agentic loop!
    builder.add_edge("tools", "web_search_agent")

    return builder.compile()



import json
from langchain_core.messages import HumanMessage
import os
import sys

root_dir = os.path.dirname(os.path.dirname(os.path.dirname(os.path.abspath(__file__))))
sys.path.append(root_dir)

from agentic_search.lib import log_if_debug
from agentic_search.graphs.web import get_search_the_web_react_graph


async def get_web_search_results(query: str):
    """
    Get a web search report for a given query using a LangGraph ReAct agent.

    Text search can be made in two ways:
    - quick search: a single search query is generated before iterative scraping
    - thorough search: multiple search queries are generated before iterative scraping
    In both cases, the results are returned as soon as the user's query is answered.

    Returns a written Markdown report of the web search result.
    """
    invocation = await get_search_the_web_react_graph().ainvoke(
        {
            "messages": [
                HumanMessage(
                    content=query
                    + f"""Answer in JSON format: {{
                        "content": "your results as a string",
                        "type": "text" | "video" (if the results are videos)
                        }}"""
                )
            ]
        }
    )
    log_if_debug(f"Web search capability result: {invocation}")
    return {
        "results": json.loads(
            invocation["messages"][-1]
            .content.replace("```json", "")
            .replace("```", "")
            .strip()
        )["content"],
        "type": json.loads(
            invocation["messages"][-1]
            .content.replace("```json", "")
            .replace("```", "")
            .strip()
        )["type"],
    }



import os
import sys
import time

from agentic_search.chains.text import (
    get_content_answers_to_query_chain,
    get_qa_summary_chain,
)
from agentic_search.functions.web import (
    get_serp_links,
    get_videos,
    get_webpages_soups_text_async,
)

root_dir = os.path.dirname(os.path.dirname(os.path.dirname(os.path.abspath(__file__))))
sys.path.append(root_dir)

from agentic_search.chains.web import (
    get_route_search_type_chain,
    get_web_search_query_chain,
    get_web_search_queries_chain,
)
from agentic_search.lib import get_websearch_llm_provider, log_if_debug


async def get_web_search_results_tool(
    query: str, estimated_number_of_searches: int, is_thorough: bool
):
    """Make a quick web search prompted by a user query and output a nicely formatted and readable Markdown document.

    Args:
        query: The input user query
        estimated_number_of_searches: Estimated number of web searches to perform to answer the user query
        is_thorough: Whether to perform each web search in thorough or quick mode (search with several search engine queries at a time VS a quick search with one search engine query at a time)

    Returns:
        str: A formatted Markdown document containing the search results

    Use this tool if you need to quickly search the web for current information or information that is not in your knowledge base.
    """
    log_if_debug(
        f"invoking web search tool with query: {query}, and {estimated_number_of_searches}, thorough: {is_thorough}"
    )

    search_type = get_route_search_type_chain().invoke({"query": query})["search_type"]

    if search_type == "video":
        vids = await get_videos(query)
        videos_with_most_views = sorted(
            vids, key=lambda x: x["statistics"]["viewCount"], reverse=True
        )
        return {
            "content": videos_with_most_views[0]["content"],
            "type": "video",
        }

    excluded_queries = []
    answer = ""
    start_time = time.time()
    for _ in range(estimated_number_of_searches):
        # check if we have an answer and if we've exceeded 30 seconds
        if answer.strip() != "" and (time.time() - start_time) > 30:
            return answer

        search_queries = (
            get_web_search_queries_chain(excluded_queries).invoke({"query": query})
            if is_thorough
            else {
                "queries": [
                    get_web_search_query_chain(excluded_queries).invoke(
                        {"query": query}
                    )["query"]
                ]
            }
        )
        for q in search_queries["queries"]:
            links_to_scrape = []
            links = await get_serp_links(q)
            if len(links) <= 0:
                continue
            links_to_scrape.extend(links)
            scraped_content = await get_webpages_soups_text_async(
                [x["href"] for x in links_to_scrape]
            )
            for item in scraped_content:
                content = answer + "\n\n" + item
                tmp_answer = get_qa_summary_chain(
                    "default", get_websearch_llm_provider()
                ).invoke(
                    {
                        "content": content,
                        "query": query,
                    }
                )
                answers_to_query = get_content_answers_to_query_chain(
                    get_websearch_llm_provider()
                ).invoke({"content": tmp_answer, "query": query})
                if answers_to_query["fully_answered"] == "yes":
                    return tmp_answer
                # re assigning `content` variable to a running summary chain on the whole content
                answer = get_qa_summary_chain(
                    "default", get_websearch_llm_provider()
                ).invoke(
                    {
                        "content": tmp_answer,
                        "query": query,
                    }
                )
                excluded_queries.append(q)
    # this always returns some form of summary, regardless of it fully being answered
    return {
        "content": get_qa_summary_chain("default", get_websearch_llm_provider()).invoke(
            {"content": answer, "query": query}
        ),
        "type": "text",
    }


def get_web_search_tools():
    return [
        get_web_search_results_tool,
    ]



from langchain_core.prompts import ChatPromptTemplate


def get_arxiv_search_queries_prompt():
    """
    Generates a prompt for generating search queries for arXiv based on a user input query.

    Returns a prompt for generating search queries.
    """
    arxiv_search_queries_prompt_messages = [
        (
            "user",
            """You are a research assistant.

Your job is to write 3 search queries to search Arxiv with a given user input query.
            
## GUIDELINES FOR QUERYING ARXIV

### searching by author:

- optimize name search by following "surname(s), forename(s) or surname(s), initial(s)" pattern: example `Hawking, S` or `Hawking, Stephen`
- for best results on multiple author names, separate individuals with a `;` (semicolon). Example: `Jin, D S; Ye, J`
- author names enclosed in quotes will return only exact matches; for example, `"Stephen Hawking"` will not return matches for `Stephen W. Hawking`
- diacritic character variants are automatically searched in the Author(s) field
- queries with no punctuation will treat each term independently

### wildcards:

- use `?` to replace a single character or `*` to replace any number of characters; this can be used in any field, but not in the first character position

### expressions:

- TeX expressions can be searched, enclosed in single `$` characters

### phrases:

- enclose phrases in double quotes for exact matches in title, abstract, and comments

### dates:

- sorting by announcement date will use the year and month the original version (v1) of the paper was announced
- sorting by submission date will use the year, month and day the latest version of the paper was submitted

### journal references:

- if a journal reference search contains a wildcard, matches will be made using wildcard matching as expected; for example, `math*` will match `math`, `maths`, `mathematics`, etc.
- if a journal reference search does not contain a wildcard, only exact phrases entered will be matched; for example, `math` would match `math` or `math` and science but not `maths` or `mathematics`
- all journal reference searches that do not contain a wildcard are literal searches: a search for `Physica A` will match all papers with journal references containing `Physica A`, but a search for `Physica A, 245 (1997) 181` will only return the paper with journal reference `Physica A, 245 (1997) 181`
       
Here is the user input query, delimited by dashes:
---
{query}
---

You must respond with a JSON containing a list of strings in the following format:
         
{{
    "queries": ["query 1", "query 2", "query 3"]
}}
         
You will ALWAYS include the original query as one of the queries.
The two other queries MUST be different from the original query.
DO NOT prefix a query with things like `title:`, `abstract:`, `author:`, etc. as this will limit the search to the specified field.""",
        )
    ]
    return ChatPromptTemplate.from_messages(arxiv_search_queries_prompt_messages)



from langchain_core.output_parsers import StrOutputParser
from langchain_core.runnables import RunnablePassthrough
import json
import os
import sys

root_dir = os.path.dirname(os.path.dirname(os.path.dirname(os.path.abspath(__file__))))
sys.path.append(root_dir)

from agentic_search.functions.arxiv import format_search_results, get_arxiv_search_results
from agentic_search.lib import get_llm
from agentic_search.prompts.arxiv import get_arxiv_search_queries_prompt
from agentic_search.prompts.text import (
    get_formatted_report_prompt,
    get_qa_summary_prompt,
)


def get_arxiv_concatenated_summary_chain():
    """
    Creates a chain that concatenates the summaries of arXiv search results.

    Input key is `query`.

    Returns a chain that concatenates the summaries of arXiv search results.
    """
    return (
        # first generate the search queries
        get_arxiv_search_queries_chain()
        # transform the queries list into a list of query objects
        | (lambda x: [{"query": q} for q in x["queries"]])
        # map each query through the search and summarization chain
        | get_arxiv_search_results_summaries_chain().map()
        # format the results into a readable formatted string
        | (lambda x: format_search_results(x))
    )


def get_arxiv_final_report_chain():
    return (
        get_formatted_report_prompt()
        | get_llm(
            "default",
            False
        )
        | StrOutputParser()
    )


def get_arxiv_search_queries_chain():
    """
    Creates a chain that generates search queries for arXiv based on a user input query.

    Input key is `query`.

    Returns a chain that generates search queries.
    """
    return (
        get_arxiv_search_queries_prompt() | get_llm() | StrOutputParser() | json.loads
    )


def get_arxiv_search_results_summaries_chain():
    """
    Creates a chain that fetches and summarizes arXiv search results.

    Returns a chain that processes arXiv search results
    """
    # 1st level: the outer `RunnablePassthrough.assign` preserves all original input and adds a 'summary' field
    summarize_arxiv_search_result_partial = RunnablePassthrough.assign(
        summary=RunnablePassthrough.assign(
            # 2nd level: takes the 'result' field from the input and assigns it to 'content'
            content=lambda input_obj: input_obj["search_result"]
        )
        # this content is then passed to the next prompt
        | get_qa_summary_prompt()
        # we don't want a JSON output here, just the raw summary string
        | get_llm(False)
        | StrOutputParser()
    )

    # this returns a list of summaries, nested in objects having `query`, `search_result`, and `summary` fields
    return (
        # from the user input query, get the search results in Arxiv
        RunnablePassthrough.assign(
            search_results=lambda input: get_arxiv_search_results(input["query"])
        )
        | (
            # transform the results into a list of dictionaries,
            # each containing a query and a search result being a formatted string
            lambda list_of_results: [
                {"query": list_of_results["query"], "search_result": result}
                for result in list_of_results["search_results"]
            ]
        )
        # we then pass each query and result to the next prompt for summarization
        | summarize_arxiv_search_result_partial.map()
    )



from langchain_core.messages import SystemMessage
from langgraph.graph import MessagesState
import os
import sys

root_dir = os.path.dirname(os.path.dirname(os.path.dirname(os.path.abspath(__file__))))
sys.path.append(root_dir)

from agentic_search.lib import get_websearch_llm, log_if_debug
from agentic_search.prompts.web import get_web_search_agent_system_prompt
from agentic_search.tools.web import get_web_search_tools


def get_web_search_agent_node(state: MessagesState):
    """Get the agent node, which is the entry point for the agent."""
    # let's give our agent a personae
    sys_msg = SystemMessage(content=get_web_search_agent_system_prompt())
    # now bind tools to the agent
    llm_with_tools = get_websearch_llm().bind_tools(get_web_search_tools())
    log_if_debug(
        f"invoking web search agent with messages: {[sys_msg] + state['messages']}"
    )
    # using messages state keeps history during the graph execution (transient to one execution)
    return {"messages": [llm_with_tools.invoke([sys_msg] + state["messages"])]}



import json
from langchain_core.output_parsers import StrOutputParser
from langchain_core.runnables import RunnablePassthrough
import os
import sys
from typing import Literal

root_dir = os.path.dirname(os.path.dirname(os.path.dirname(os.path.abspath(__file__))))
sys.path.append(root_dir)

from agentic_search.functions.text import get_pdf_pages_docs
from agentic_search.lib import get_llm
from agentic_search.prompts.text import (
    get_claims_consistency_comparison_prompt,
    get_content_answers_to_query_prompt,
    get_content_summary_prompt,
    get_formatted_report_prompt,
    get_qa_summary_prompt,
)


def get_claims_consistency_comparison_chain():
    return (
        get_claims_consistency_comparison_prompt()
        | get_llm()
        | StrOutputParser()
        | json.loads
    )


def get_content_answers_to_query_chain(llm_provider: Literal["ollama", "openai"] = "ollama"):
    return (
        get_content_answers_to_query_prompt()
        | get_llm("default", True, llm_provider)
        | StrOutputParser()
        | json.loads
    )


def get_content_summary_chain(llm_provider: Literal["ollama", "openai"] = "ollama"):
    return get_content_summary_prompt() | get_llm("default", True, llm_provider) | StrOutputParser()


def get_pdf_report_chain():
    """
    Generates a report chain for a PDF document.

    Input key is `source`.
    """
    return (
        RunnablePassthrough.assign(
            results=lambda input: get_pdf_pages_docs(input["source"])
        )
        | (
            lambda pages: [
                {
                    "content": page.page_content,
                }
                for page in pages["results"]
            ]
        )
        | get_qa_summary_chain().map()
        | (
            lambda summaries: {
                "unstructured_text": "\n\n".join([f"""{s}""" for s in summaries])
            }
        )
        | RunnablePassthrough.assign(unstructured_text=lambda input: input)
        | get_formatted_report_prompt()
        | get_llm("long-context", False)
        | StrOutputParser()
    )


def get_qa_summary_chain(use_case: Literal["default", "long-context"] = "default", llm_provider: Literal["ollama", "openai"] = "ollama"):
    """
    Generates a summary chain.

    Input keys are `content` and `query`.
    """
    return get_qa_summary_prompt(True) | get_llm(use_case, False, llm_provider) | StrOutputParser()



from aiocache import Cache
import asyncio
from asyncio import Semaphore
from bs4 import BeautifulSoup
from concurrent.futures import ThreadPoolExecutor
from duckduckgo_search import DDGS
from duckduckgo_search.exceptions import DuckDuckGoSearchException, RatelimitException
import json
import os
from selenium import webdriver
from selenium.webdriver.chrome.options import Options
from selenium.webdriver.support.ui import WebDriverWait
import sys
from typing import List, Literal

root_dir = os.path.dirname(os.path.dirname(os.path.dirname(os.path.abspath(__file__))))
sys.path.append(root_dir)

from agentic_search.lib import log_if_debug

# sharing only one Chrome instance across all requests
chrome_semaphore = Semaphore(1)


class CachedDDGWrapper:
    def __init__(self, max_results: int):
        use_redis = os.getenv("USE_REDIS", "false").lower() == "true"
        if use_redis:
            self.cache = Cache(Cache.REDIS, ttl=300)  # 5-minute redis cache
        else:
            self.cache = Cache(Cache.MEMORY, ttl=300)  # 5-minute in memory cache
        self.max_results = max_results

    async def results(self, query, type: Literal["text", "video"] = "text"):
        # check cache first
        cached_result = await self.cache.get(query)
        if cached_result:
            return json.loads(cached_result)

        # simplified retry logic - only one retry
        ddg_wrapper = None
        try:
            ddg_wrapper = DDGS(proxy=None)
            results = (
                ddg_wrapper.text(query, max_results=self.max_results)
                if type == "text"
                else ddg_wrapper.videos(query, max_results=self.max_results)
            )

            if not results:
                raise DuckDuckGoSearchException("No results returned")

            await self.cache.set(query, json.dumps(results))
            return results

        except (DuckDuckGoSearchException, RatelimitException, Exception) as e:
            log_if_debug(f"Search attempt failed: {str(e)}")
            return []


def get_chrome_instance():
    chrome_options = Options()
    chrome_options.add_argument("--headless=new")
    chrome_options.add_argument("--no-sandbox")
    chrome_options.add_argument("--remote-debugging-port=9222")
    chrome_options.add_argument("--window-size=1920,1080")
    if os.getenv("RUNS_ON_DOCKER", "false").lower() == "true":
        chrome_options.add_argument("--disable-gpu")
        chrome_options.add_argument("--disable-dev-shm-usage")
    chrome_options.binary_location = os.getenv(
        "CHROME_BINARY_LOCATION", "/usr/bin/google-chrome"
    )
    chrome_options.page_load_strategy = "eager"

    # the Chrome driver is the "bridge" between Python code and the Chrome browser;
    # the service starts and manages the chrome driver binary that actually controls Chrome,
    # it also handles the communication between Python and the driver,
    # the provided timeout is the timeout for the service to start the driver
    chrome_instance = webdriver.Chrome(
        options=chrome_options,
        service=webdriver.ChromeService(timeout=2),
    )
    return chrome_instance


async def get_serp_links(query: str, num_results: int = 3):
    ddg_search = CachedDDGWrapper(max_results=num_results)
    results = await ddg_search.results(query)
    log_if_debug(f"serp results for query {query}: {results}")
    return results


async def get_videos(query: str, num_results: int = 3):
    ddg_search = CachedDDGWrapper(max_results=num_results)
    results = await ddg_search.results(query, "video")
    log_if_debug(f"serp results for query {query}: {results}")
    return results


def get_webpage_soup(
    webpage_url: str, chrome_instance: webdriver.Chrome, timeout: int = 4
) -> BeautifulSoup:
    try:
        wait = WebDriverWait(chrome_instance, timeout)
        chrome_instance.get(webpage_url)

        # check if browser is still responsive
        try:
            chrome_instance.current_url  # this will throw if browser crashed
        except:
            # clean up crashed instance
            chrome_instance.quit()
            raise Exception("browser became unresponsive")

        wait.until(
            lambda d: d.execute_script("return document.readyState") == "complete",
        )
        return BeautifulSoup(chrome_instance.page_source, "html.parser")
    except Exception as e:
        log_if_debug(
            f"`get_webpage_soup` => error getting webpage soup for {webpage_url}: {e}"
        )
        return None


def get_webpage_soup_text(
    webpage_url: str, chrome_instance: webdriver.Chrome, timeout: int = 4
) -> str:
    soup = None
    try:
        soup = get_webpage_soup(webpage_url, chrome_instance, timeout)
        if soup is None:
            return ""
        text = soup.get_text(separator=" ", strip=True)
        text += f"\n\nSOURCE: {webpage_url}\n\n"
        return text
    except Exception as e:
        log_if_debug(
            f"`get_webpage_soup_text` => error getting webpage soup for {webpage_url}: {e}"
        )
        return ""


async def get_webpage_soup_text_async(webpage_url: str, timeout: int = 4) -> str:
    async with chrome_semaphore:  # wait for available slot
        # create a new Chrome instance for each request to avoid thread safety issues
        chrome_instance = get_chrome_instance()
        try:
            # run the synchronous selenium operations in a thread pool
            loop = asyncio.get_running_loop()
            with ThreadPoolExecutor() as pool:
                text = await loop.run_in_executor(
                    pool,
                    lambda: get_webpage_soup_text(
                        webpage_url, chrome_instance, timeout
                    ),
                )
                # always close the Chrome instance
                chrome_instance.quit()
                return text
        except Exception as e:
            log_if_debug(
                f"`get_webpage_soup_text_async` => error getting webpage soup for {webpage_url}: {e}"
            )
            chrome_instance.quit()
            return ""


async def get_webpages_soups_text_async(
    urls: List[str], timeout_for_page_load: int = 3
) -> List[str]:
    # create tasks for all URLs
    tasks = [get_webpage_soup_text_async(url, timeout_for_page_load) for url in urls]

    # gather all results,
    # using `return_exceptions=True` to return exceptions as well;
    # this allows to not stop the scraping process when one page fails to load
    soups_text = await asyncio.gather(*tasks, return_exceptions=True)

    # filter out exceptions and combine text
    content = []
    for text in soups_text:
        if isinstance(text, Exception):
            log_if_debug(f"error getting webpages soups text: {text}")
            continue
        if text:  # Check if text is not None
            content.append(text)

    return content
